; Tracer v2.0 - utility to trace int 21h calls
; Written by Tom Torfs (2:292/516@fidonet.org, tomtorfs@mail.dma.be)
; Donated to the public domain, 1997-08-03
; v2.0 - gives extended information, 1999-03-21

; extra information id bytes (function number+desc is always printed)
i_end equ 0    ; end of extra information list
; registers
r_ax  equ 1    ; normal registers
r_bx  equ 2
r_cx  equ 3
r_dx  equ 4
r_es  equ 5
r_dxax equ 6   ; register combinations (dx:ax etc.)
r_cxdx equ 7
r_sidi equ 8
r_dsdx equ 9
r_dsbx equ 10
r_esbx equ 11
r_dssi equ 12
r_esdi equ 13
r_cflag equ 14 ; carry flag
; formatting types (strings also print addresses)
f_32 equ 1     ; 32-bit hex number (for register combinations)
f_16 equ 2     ; 16-bit hex number (for full registers)
f_l8 equ 3     ; low 8-bits hex number (for bottom 8-bit registers al etc.)
f_h8 equ 4     ; high 8-bits hex number (for top 8-bit registers ah etc.)
f_ch equ 5     ; low 8-bits character (hex if <32 or >126)
f_ch0 equ 6    ; low 8-bits character 'A'..'Z' with A=0
f_ch1 equ 7    ; low 8-bits character 'A'..'Z' with A=1 (0 will become ?)
f_str$ equ 8   ; dollar-terminated string
f_str0 equ 9   ; 0-terminated string
f_str2 equ 10  ; string with length in byte 0, data starting at byte 2
f_stra equ 11  ; raw binary string; must be followed by # of char
               ; (0 = value taken from cx, 255 = if CF=0 AX, otherwise nothing)
f_strb equ 12  ; raw binary string; must be followed by # of char
               ; (0 = value taken from cx, 255 = if CF=0 AX, otherwise nothing)
f_fcb equ 13   ; FCB filename (filename.ext)
f_bool equ 14  ; 0=OK, nonzero=ERROR

program segment para public 'code'
assume cs:program,ds:program,es:program,ss:program

org 100h
@@start:

@@codestart:

; make the stack not bigger than necessary
mov sp,offset @@stackstart+512

; free all memory we don't need
mov ah,4Ah
mov bx,sp
add bx,15
mov cl,4
shr bx,cl
int 21h

; show title
mov ah,9
mov dx,offset titlemessage
int 21h

; skip leading spaces
mov si,81h
@@skipspaceloop:
   cmp byte ptr [si],' '
   jne @@skipspacedone
   inc si
jmp @@skipspaceloop
@@skipspacedone:

; check for no parameters
cmp byte ptr [si],0Dh
jne @@cmdlineok
   ; show usage info
   mov ah,9
   mov dx,offset usage
   int 21h
   ; terminate
   mov ax,4C01h
   int 21h
@@cmdlineok:

; nul-terminate executable name
mov bx,si
@@nultermloop:
   cmp byte ptr [bx],' '
   je @@nultermdone
   cmp byte ptr [bx],0Dh
   je @@nultermdone
   inc bx
jmp @@nultermloop
@@nultermdone:
mov dl,0
xchg [bx],dl

; expand executable name using DOS canonicalize function
mov ah,60h
mov di,offset execpath
int 21h

; restore byte under nulbyte
xchg [bx],dl

; calculate number of characters to cut off
mov ax,bx
sub ax,81h

; adjust commandline length
mov cl,ds:[80h]
mov ch,0
sub cx,ax
mov ds:[80h],cl

; move commandline over program name
cld
mov si,bx
mov di,81h
inc cx   ; include closing 0Dh
rep movsb

; show tracing message
mov ah,9
mov dx,offset tracing
int 21h
mov dx,offset execpath
call @@puts

; open logfile for writing
mov ah,3Ch
xor cx,cx
mov dx,offset logfile
int 21h
jnc @@logopenok
   ; show error message
   mov ah,9
   mov dx,offset openlogerr
   int 21h
   ; terminate
   mov ax,4C01h
   int 21h
@@logopenok:
mov [loghandle],ax

; store our PSP address
mov ah,62h
int 21h
mov [ourpsp],bx

; store original int 21h handler
mov ax,3521h
int 21h
mov [oldint21ofs],bx
mov [oldint21seg],es
mov ax,cs
mov es,ax

; install our int 21h handler
mov ax,2521h
mov dx,offset @@int21handler
int 21h

; fill in parameter block
mov [commandlineseg],ds
mov [commandlineofs],80h
mov ax,ds:[2Ch]
mov [envseg],ax

; save stack pointer
mov [saveSP],sp

; execute program
mov ax,4B00h
mov bx,offset parablock
mov dx,offset execpath
int 21h

; restore segment registers & stack pointer
mov ax,cs
mov ds,ax
mov es,ax
mov ss,ax
mov sp,[saveSP]

jnc @@noerror
   ; show error message
   mov ah,9
   mov dx,offset execerror
   int 21h
@@noerror:

; restore old int 21h handler
mov ax,2521h
mov dx,[oldint21ofs]
mov ds,[oldint21seg]
int 21h
mov ax,cs
mov ds,ax

; close logfile
mov ah,3Eh
mov bx,[loghandle]
int 21h

; terminate
mov ax,4C00h
int 21h

; int 21h handler
@@int21handler:
   push bp
   mov bp,sp
   push ax     ; bp-2
   push bx     ; bp-4
   push cx     ; bp-6
   push dx     ; bp-8
   push ds     ; bp-10
   pushf       ; bp-12
   mov ax,cs
   mov ds,ax
   cmp [indos],0
   je @@intercept
   jmp @@nointercept    ; recursive call
@@intercept:
   ; get PSP address
   mov ah,62h
   call @@calloldint21
   cmp bx,[ourpsp]
   jne @@intercept2
   jmp @@nointercept    ; called from within tracer program
@@intercept2:
   ; fetch registers
   call @@fetchregisters
   ; get segment address
   mov dx,[m_cs]
   ; calculate segment address relative to PSP
   sub dx,bx
   jnc @@intercept3
   jmp @@nointercept    ; called from outside program space
@@intercept3:
   ; signal we're active to be aware of recursive calls
   mov [indos],1
   ; get offset address
   mov ax,[m_ip]
   ; substract 2 bytes for INT 21h
   sub ax,2
   ; write address in DX:AX
   call @@writehex32
   ; write space
   mov dx,offset space
   mov cx,1
   call @@writelog
   ; write AX=
   mov dx,offset axvalue
   mov cx,3
   call @@writelog
   ; write ax value
   mov ax,[m_ax]
   call @@writehex16
   mov [xinfopos],0
   cmp byte ptr [m_ax+1],73h     ; ah
   ja @@nodesc
      ; write space
      mov dx,offset space
      mov cx,1
      call @@writelog
      ; write description
      mov bl,byte ptr [m_ax+1]   ; ah
      mov bh,0
      shl bx,1
      mov dx,[funcdesc+bx]
      mov [xinfopos],dx
      mov cl,0
      push es
      mov ax,ds
      mov es,ax
      mov bx,dx
      call @@writestr
      pop es
      inc ax                  ; ax=number of characters excl. nul byte
      add [xinfopos],ax
   @@nodesc:
   ; write crlf
   mov dx,offset crlf
   mov cx,2
   call @@writelog
   ; write extra info (if any)
      call @@writeinfo
   popf
   pop ds
   pop dx
   pop cx
   pop bx
   pop ax
   pop bp
   call @@calloldint21
   push bp
   mov bp,sp
   push ax     ; bp-2
   push bx     ; bp-4
   push cx     ; bp-6
   push dx     ; bp-8
   push ds     ; bp-10
   pushf       ; bp-12
   mov ax,cs
   mov ds,ax
   ; fetch registers
   call @@fetchregisters
   ; write extra info (if any)
   call @@writeinfo
   mov [indos],0
   pop ax   ; fetch flags
   or ax,[m_cflag]   ; add carry flag result
   push ax
   popf     ; restore flags
   pop ds
   pop dx
   pop cx
   pop bx
   pop ax
   pop bp
   iret
@@nointercept:
   popf
   pop ds
   pop dx
   pop cx
   pop bx
   pop ax
   pop bp
   db 0EAh  ; jmp seg:ofs
oldint21addr label dword
   oldint21ofs dw ?
   oldint21seg dw ?

@@calloldint21:
   pushf
   call dword ptr cs:[oldint21addr]
   ret

; fetches registers
; ax,bx,cx,dx,ds,cflag and caller cs:ip are fetched from stack
; si, di and es are fetched from registers
; changes ax
@@fetchregisters:
   mov ax,[bp-2]
   mov [m_ax],ax
   mov ax,[bp-4]
   mov [m_bx],ax
   mov ax,[bp-6]
   mov [m_cx],ax
   mov ax,[bp-8]
   mov [m_dx],ax
   mov [m_si],si
   mov [m_di],di
   mov ax,[bp-10]
   mov [m_ds],ax
   mov [m_es],es
   mov ax,[bp+4]
   mov [m_cs],ax
   mov ax,[bp+2]
   mov [m_ip],ax
   mov ax,[bp-12]
   and ax,1
   mov [m_cflag],ax
   ret

; prints a nul-terminated string, followed by a crlf
; ds:dx = string
; changes: ax,bx,dx
@@puts:
   mov bx,dx
   dec bx ; to compensate for first inc
   @@putszeroloop:
      inc bx
      cmp byte ptr [bx],0
   jne @@putszeroloop
   mov byte ptr [bx],'$'
   mov ah,9
   int 21h
   mov byte ptr [bx],0
   mov ah,9
   mov dx,offset crlf
   int 21h
   ret

; write cx bytes from ds:dx to the logfile
; changes ax,bx,cx
@@writelog:
   mov ah,40h
   mov bx,cs:[loghandle]      ; cs: override to allow different ds
   int 21h
   ret

; converts the 4-digit hexadecimal value in ax to hexbuf
; changes bx,cx
@@convhex:
   mov bl,ah
   mov cl,4
   shr bl,cl
   mov bh,0
   mov cl,[hexdigit+bx]
   mov [hexbuf],cl
   mov bl,ah
   and bl,0Fh
   mov bh,0
   mov cl,[hexdigit+bx]
   mov [hexbuf+1],cl
   mov bl,al
   mov cl,4
   shr bl,cl
   mov bh,0
   mov cl,[hexdigit+bx]
   mov [hexbuf+2],cl
   mov bl,al
   and bl,0Fh
   mov bh,0
   mov cl,[hexdigit+bx]
   mov [hexbuf+3],cl
   ret

; changes ax,bx,cx,dx
@@writehex16:
   call @@convhex
   mov dx,offset hexbuf
   mov cx,4
   call @@writelog
   ret

; write the 2-digit hexadecimal value in al to the logfile
; changes ax,bx,cx,dx
@@writehex8:
   mov ah,0
   call @@convhex
   mov dx,offset hexbuf + 2
   mov cx,2
   call @@writelog
   ret

; write the 8-digit hexadecimal value in dx:ax to the logfile
; changes ax,bx,cx,dx
@@writehex32:
   push ax
   mov ax,dx
   call @@writehex16
   mov dx,offset colon
   mov cx,1
   call @@writelog
   pop ax
   call @@writehex16
   ret

; write the 8-bit character in al to the logfile if printable
; (writes hexadecimal value otherwise)
; changes ax,bx,cx,dx
@@writech:
   cmp al,32
   jb @@notprintable
   cmp al,126
   ja @@notprintable
   mov byte ptr [hexbuf],al
   mov dx,offset hexbuf
   mov cx,1
   call @@writelog
   ret
@@notprintable:
   push ax
   mov byte ptr [hexbuf],'<'
   mov dx,offset hexbuf
   mov cx,1
   call @@writelog
   pop ax
   call @@writehex8
   mov byte ptr [hexbuf],'>'
   mov dx,offset hexbuf
   mov cx,1
   call @@writelog
   ret

; write a string of upto cx characters pointed to by es:bx to the logfile;
; unprintable characters are replaced by .
; changes ax,bx,cx
@@writerawstr:
@@writerawstrloop:
      mov al,es:[bx]
      inc bx
      push bx
      push cx
      call @@writech
      pop cx
      pop bx
   loop @@writerawstrloop
   ret

; write a character cl terminated string from es:bx to the logfile
; changes ax,bx,cx,dx
; returns string length (excluding terminating byte) in ax
@@writestr:
   mov dx,bx
   dec bx   ; compensate for first inc
   @@findstrzeroloop:
      inc bx
      cmp byte ptr [bx],cl
   jnz @@findstrzeroloop
   mov cx,bx
   sub cx,dx
   push cx
   mov bx,dx
   call @@writerawstr
   pop ax
   ret

; write a string of upto cx characters pointed to by es:bx to the logfile;
; characters are printed in hexadecimal form, separated by spaces
; changes ax,bx,cx
@@writebinstr:
@@writebinstrloop:
      mov al,es:[bx]
      inc bx
      push bx
      push cx
      call @@writehex8
      mov dx,offset space
      mov cx,1
      call @@writelog
      pop cx
      pop bx
   loop @@writebinstrloop
   ret

; write filename from the FCB pointed to by ds:dx to the logfile
; changes ax,bx,cx
@@writefcb:
   ; write filename
   inc dx   ; skip device (or FF for XFCB)
   mov cx,1
   call @@writelog
   ; write dot
   push dx
   push ds
   mov ax,cs
   mov ds,ax
   mov dx,offset dot
   mov cx,1
   call @@writelog
   pop ds
   pop dx
   ; write extension
   add dx,8
   mov cx,1
   call @@writelog
   ret

; write OK if AX==0 or ERROR if AX!=0
; changes ax,bx,cx,dx
@@writebool:
   test ax,ax
   jz @@boolok
   ; write 'ERROR'
   mov dx,offset error
   mov cx,5
   call @@writelog
   ret
@@boolok:
   ; write 'OK'
   mov dx,offset ok
   mov cx,2
   call @@writelog
   ret

; writes extra information (if any)
; changes ax,bx,cx,dx
@@writeinfo:
   cmp [xinfopos],0
   jnz @@exinfoloop
   ret
@@exinfoloop:
   mov bx,[xinfopos]
   mov cl,ds:[bx]
   inc [xinfopos]
   cmp cl,i_end
   jne @@exinfonotend
      ret
   @@exinfonotend:
   xor dx,dx   ; dx:ax will contain saved value
   xor ax,ax
   cmp cl,r_ax
   jne @@notsaveax
      mov ax,[m_ax]
      jmp @@exinfoformat
@@notsaveax:
   cmp cl,r_bx
   jne @@notsavebx
      mov ax,[m_bx]
      jmp @@exinfoformat
@@notsavebx:
   cmp cl,r_cx
   jne @@notsavecx
      mov ax,[m_cx]
      jmp @@exinfoformat
@@notsavecx:
   cmp cl,r_dx
   jne @@notsavedx
      mov ax,[m_dx]
      jmp @@exinfoformat
@@notsavedx:
   cmp cl,r_es
   jne @@notsavees
      mov ax,[m_es]
      jmp @@exinfoformat
@@notsavees:
   cmp cl,r_dxax
   jne @@notsavedxax
      mov dx,[m_dx]
      mov ax,[m_ax]
      jmp @@exinfoformat
@@notsavedxax:
   cmp cl,r_cxdx
   jne @@notsavecxdx
      mov dx,[m_cx]
      mov ax,[m_dx]
      jmp @@exinfoformat
@@notsavecxdx:
   cmp cl,r_sidi
   jne @@notsavesidi
      mov dx,[m_si]
      mov ax,[m_di]
      jmp @@exinfoformat
@@notsavesidi:
   cmp cl,r_dsdx
   jne @@notsavedsdx
      mov dx,[m_ds]
      mov ax,[m_dx]
      jmp @@exinfoformat
@@notsavedsdx:
   cmp cl,r_esbx
   jne @@notsaveesbx
      mov dx,[m_es]
      mov ax,[m_bx]
      jmp @@exinfoformat
@@notsaveesbx:
   cmp cl,r_dssi
   jne @@notsavedssi
      mov dx,[m_ds]
      mov ax,[m_si]
      jmp @@exinfoformat
@@notsavedssi:
   cmp cl,r_esdi
   jne @@notsaveesdi
      mov dx,[m_es]
      mov ax,[m_di]
      jmp @@exinfoformat
@@notsaveesdi:
   cmp cl,r_cflag
   jne @@notsavecflag
      mov ax,[m_cflag]
      jmp @@exinfoformat
@@notsavecflag:
   ; unknown value, abort
   ret
@@exinfoformat:
   ; save dx:ax
   push dx
   push ax
   ; fetch format byte
   mov bx,[xinfopos]
   mov cl,ds:[bx]
   inc bx
   mov ch,0
   cmp cl,f_stra
   je @@getstrsize
   cmp cl,f_strb
   jne @@nostrsize
@@getstrsize:
      mov ch,ds:[bx]
      inc bx
   @@nostrsize:
   mov [xinfopos],bx
   push cx  ; save format byte (& string size info)
   ; write spaces
   mov dx,offset space
   mov cx,3
   call @@writelog
   ; write description
   mov cl,0
   push es
   mov ax,ds
   mov es,ax
   mov bx,[xinfopos]
   call @@writestr
   pop es
   inc ax      ; number of characters written excl. 0 byte
   add [xinfopos],ax
   ; write colon
   mov dx,offset colon
   mov cx,1
   call @@writelog
   ; write space
   mov dx,offset space
   mov cx,1
   call @@writelog
   ; restore format byte
   pop cx
   ; restore dx:ax
   pop ax
   pop dx
   ; check format byte
   cmp cl,f_32
   jne @@notformat32
      call @@writehex32
      jmp @@exinfoformatdone
@@notformat32:
   cmp cl,f_16
   jne @@notformat16
      call @@writehex16
      jmp @@exinfoformatdone
@@notformat16:
   cmp cl,f_l8
   jne @@notformatl8
      call @@writehex8
      jmp @@exinfoformatdone
@@notformatl8:
   cmp cl,f_h8
   jne @@notformath8
      mov al,ah
      call @@writehex8
      jmp @@exinfoformatdone
@@notformath8:
   cmp cl,f_ch
   jne @@notformatch
      call @@writech
      jmp @@exinfoformatdone
@@notformatch:
   cmp cl,f_ch0
   jne @@notformatch0
      add al,'A'
      cmp al,'Z'
      jna @@ch0notoverz
         mov al,'?'
      @@ch0notoverz:
      call @@writech
      jmp @@exinfoformatdone
@@notformatch0:
   cmp cl,f_ch1
   jne @@notformatch1
      add al,'A'-1
      cmp al,'Z'
      jna @@ch1notoverz
         mov al,'?'
      @@ch1notoverz:
      call @@writech
      jmp @@exinfoformatdone
@@notformatch1:
   cmp cl,f_str$
   jne @@notformatstr$
      push ax
      push dx
      call @@writehex32
      mov dx,offset space
      mov cx,1
      call @@writelog
      pop dx
      pop ax
      push es
      mov es,dx
      mov bx,ax
      mov cl,'$'
      call @@writestr
      pop es
      jmp @@exinfoformatdone
@@notformatstr$:
   cmp cl,f_str0
   jne @@notformatstr0
      push ax
      push dx
      call @@writehex32
      mov dx,offset space
      mov cx,1
      call @@writelog
      pop dx
      pop ax
      push es
      mov es,dx
      mov bx,ax
      mov cl,0
      call @@writestr
      pop es
      jmp @@exinfoformatdone
@@notformatstr0:
   cmp cl,f_str2
   jne @@notformatstr2
      push ax
      push dx
      call @@writehex32
      mov dx,offset space
      mov cx,1
      call @@writelog
      pop dx
      pop ax
      push ds
      mov ds,dx
      mov dx,ax
      mov bx,dx
      add dx,2
      mov cl,ds:[bx]
      mov ch,0
      call @@writelog
      pop ds
      jmp @@exinfoformatdone
@@notformatstr2:
   cmp cl,f_stra
   jne @@notformatstra
      push ax
      push cx
      push dx
      call @@writehex32
      mov dx,offset space
      mov cx,1
      call @@writelog
      pop dx
      pop cx
      pop ax
      push es
      mov es,dx
      mov bx,ax
      mov cl,ch
      mov ch,0
      cmp cl,255
      jne @@stranotax
         cmp [m_cflag],0
         jne @@skipstra
         mov cx,[m_ax]
         jmp @@stranotcx
      @@stranotax:
      test cx,cx
      jnz @@stranotcx
         mov cx,[m_cx]
      @@stranotcx:
      call @@writerawstr
   @@skipstra:
      pop es
      jmp @@exinfoformatdone
@@notformatstra:
   cmp cl,f_strb
   jne @@notformatstrb
      push ax
      push cx
      push dx
      call @@writehex32
      mov dx,offset space
      mov cx,1
      call @@writelog
      pop dx
      pop cx
      pop ax
      push es
      mov es,dx
      mov bx,ax
      mov cl,ch
      mov ch,0
      cmp cl,255
      jne @@strbnotax
         cmp [m_cflag],0
         jne @@skipstrb
         mov cx,[m_ax]
         jmp @@strbnotcx
      @@strbnotax:
      test cx,cx
      jnz @@strbnotcx
         mov cx,[m_cx]
      @@strbnotcx:
      call @@writebinstr
   @@skipstrb:
      pop es
      jmp @@exinfoformatdone
@@notformatstrb:
   cmp cl,f_fcb
   jne @@notformatfcb
      push ax
      push dx
      call @@writehex32
      mov dx,offset space
      mov cx,1
      call @@writelog
      pop dx
      pop ax
      push ds
      mov ds,dx
      mov dx,ax
      call @@writefcb
      pop ds
      jmp @@exinfoformatdone
@@notformatfcb:
   cmp cl,f_bool
   jne @@exinfoformatdone
      call @@writebool
@@exinfoformatdone:
   ; write crlf
   mov dx,offset crlf
   mov cx,2
   call @@writelog
jmp @@exinfoloop

parablock label
envseg dw ?
commandlineofs dw ?
commandlineseg dw ?
fcb1ofs dw -1
fcb1seg dw -1
fcb2ofs dw -1
fcb2seg dw -1

saveSP dw ?

indos db 0

hexdigit db '0123456789ABCDEF'

hexbuf db 4 dup (?)

titlemessage db 'Tracer v2.0, public domain by Tom Torfs',13,10,10,'$'

usage db 'Usage: TRACER <executable> [<commandline parameters>]',13,10,10
      db 'The trace data will be written to TRACER.LOG.',13,10
      db '$'

tracing db 'Tracing ','$'

execerror db 'Error executing program',13,10,'$'

openlogerr db 'Can''t write to logfile',13,10,'$'

logfile db 'TRACER.LOG',0

colon    db ':'
dot      db '.'
space    db '   '
axvalue  db 'AX='
ok       db 'OK'
error    db 'ERROR'
crlf     db 13,10,'$'

m_ax dw ?
m_bx dw ?
m_cx dw ?
m_dx dw ?
m_si dw ?
m_di dw ?
m_ds dw ?
m_es dw ?
m_cs dw ?
m_ip dw ?
m_cflag dw ?

xinfopos dw ?

loghandle dw ?

ourpsp dw ?

execpath db 80 dup (?)

funcdesc dw offset func00, offset func01, offset func02, offset func03
         dw offset func04, offset func05, offset func06, offset func07
         dw offset func08, offset func09, offset func0A, offset func0B
         dw offset func0C, offset func0D, offset func0E, offset func0F
         dw offset func10, offset func11, offset func12, offset func13
         dw offset func14, offset func15, offset func16, offset func17
         dw offset func18, offset func19, offset func1A, offset func1B
         dw offset func1C, offset func1D, offset func1E, offset func1F
         dw offset func20, offset func21, offset func22, offset func23
         dw offset func24, offset func25, offset func26, offset func27
         dw offset func28, offset func29, offset func2A, offset func2B
         dw offset func2C, offset func2D, offset func2E, offset func2F
         dw offset func30, offset func31, offset func32, offset func33
         dw offset func34, offset func35, offset func36, offset func37
         dw offset func38, offset func39, offset func3A, offset func3B
         dw offset func3C, offset func3D, offset func3E, offset func3F
         dw offset func40, offset func41, offset func42, offset func43
         dw offset func44, offset func45, offset func46, offset func47
         dw offset func48, offset func49, offset func4A, offset func4B
         dw offset func4C, offset func4D, offset func4E, offset func4F
         dw offset func50, offset func51, offset func52, offset func53
         dw offset func54, offset func55, offset func56, offset func57
         dw offset func58, offset func59, offset func5A, offset func5B
         dw offset func5C, offset func5D, offset func5E, offset func5F
         dw offset func60, offset func61, offset func62, offset func63
         dw offset func64, offset func65, offset func66, offset func67
         dw offset func68, offset func69, offset func6A, offset func6B
         dw offset func6C, offset func6D, offset func6E, offset func6F
         dw offset func70, offset func71, offset func72, offset func73


; for each function, the 0-terminated description string is specified
; after that a list of extra pre-call information can be specified,
; consisting of an identification byte (see above) followed by a
; 0-terminated description, i_endlist ends; then a list of extra
; post-call information can follow, again i_endlist ends

func00 db "terminate program",0
       db i_end
       db i_end
func01 db "character input with echo",0
       db i_end
       db r_ax,f_ch,"ch",0
       db i_end
func02 db "character output",0
       db r_dx,f_ch,"ch",0
       db i_end
       db i_end
func03 db "character input from AUX",0
       db i_end
       db r_ax,f_ch,"ch",0
       db i_end
func04 db "character output to AUX",0
       db r_dx,f_ch,"ch",0
       db i_end
       db i_end
func05 db "character output to PRN",0
       db r_dx,f_ch,"ch",0
       db i_end
       db i_end
func06 db "direct console I/O",0
       db r_dx,f_ch,"ch (FF=read)",0
       db i_end
       db r_ax,f_ch,"ch read (if FF)",0
       db i_end
func07 db "unfiltered character input without echo",0
       db i_end
       db r_ax,f_ch,"ch",0
       db i_end
func08 db "character input without echo",0
       db i_end
       db r_ax,f_ch,"ch",0
       db i_end
func09 db "output character string",0
       db r_dsdx,f_str$,"str",0
       db i_end
       db i_end
func0A db "buffered input",0
       db i_end
       db r_dsdx,f_str2,"str",0
       db i_end
func0B db "get input status (character ready?)",0
       db i_end
       db r_ax,f_l8,"status",0
       db i_end
func0C db "reset input buffer and then input",0
       db r_ax,f_l8,"func#",0
       db r_dsdx,f_32,"buffer",0
       db i_end
       db r_ax,f_ch,"ch (if not 0A)",0
       db i_end
func0D db "disk reset",0
       db i_end
       db i_end
func0E db "select default disk drive",0
       db r_dx,f_ch0,"drive",0
       db i_end
       db i_end
func0F db "open file (FCB)",0
       db r_dsdx,f_fcb,"name",0
       db i_end
       db r_ax,f_l8,"status",0
       db i_end
func10 db "close file (FCB)",0
       db r_dsdx,f_fcb,"name",0
       db i_end
       db r_ax,f_l8,"status",0
       db i_end
func11 db "search for first match (FCB)",0
       db r_dsdx,f_fcb,"mask",0
       db i_end
       db r_ax,f_l8,"status",0
       db i_end
func12 db "search for next match (FCB)",0
       db r_dsdx,f_fcb,"mask",0
       db i_end
       db r_ax,f_l8,"status",0
       db i_end
func13 db "delete file (FCB)",0
       db r_dsdx,f_fcb,"name",0
       db i_end
       db r_ax,f_l8,"status",0
       db i_end
func14 db "sequential read (FCB)",0
       db r_dsdx,f_fcb,"name",0
       db i_end
       db r_ax,f_l8,"status",0
       db i_end
func15 db "sequential write (FCB)",0
       db r_dsdx,f_fcb,"name",0
       db i_end
       db r_ax,f_l8,"status",0
       db i_end
func16 db "create file (FCB)",0
       db r_dsdx,f_fcb,"name",0
       db i_end
       db r_ax,f_l8,"status",0
       db i_end
func17 db "rename file (FCB)",0
       db r_dsdx,f_fcb,"name",0
       db i_end
       db r_ax,f_l8,"status",0
       db i_end
func18 db "CP/M compatability function",0
       db i_end
       db i_end
func19 db "get default disk drive",0
       db i_end
       db r_ax,f_ch0,"drive",0
       db i_end
func1A db "set DTA address",0
       db r_dsdx,f_32,"address",0
       db i_end
       db i_end
func1B db "get information for default drive",0
       db i_end
       db r_ax,f_l8,"sectors/cluster",0
       db r_dx,f_16,"# of clusters",0
       db r_dsbx,f_strb,1,"media descriptor",0
       db i_end
func1C db "get information for specified drive",0
       db r_dx,f_ch1,"drive",0
       db i_end
       db r_ax,f_l8,"sectors/cluster",0
       db r_dx,f_16,"# of clusters",0
       db r_dsbx,f_strb,1,"media descriptor",0
       db i_end
func1D db "CP/M compatability function",0
       db i_end
       db i_end
func1E db "CP/M compatability function",0
       db i_end
       db i_end
func1F db "get DPB pointer to current drive",0
       db i_end
       db r_ax,f_l8,"status",0
       db r_dsbx,f_strb,33,"DPB",0
       db i_end
func20 db "CP/M compatability function",0
       db i_end
       db i_end
func21 db "random read (FCB)",0
       db r_dsdx,f_fcb,"name",0
       db i_end
       db r_ax,f_l8,"status",0
       db i_end
func22 db "random write (FCB)",0
       db r_dsdx,f_fcb,"name",0
       db i_end
       db r_ax,f_l8,"status",0
       db i_end
func23 db "get file size in records (FCB)",0
       db r_dsdx,f_fcb,"name",0
       db i_end
       db r_ax,f_l8,"#records",0
       db i_end
func24 db "set random record number (FCB)",0
       db r_dsdx,f_fcb,"name",0
       db i_end
func25 db "set interrupt vector",0
       db r_ax,f_l8,"int#",0
       db r_dsdx,f_32,"address",0
       db i_end
       db i_end
func26 db "create PSP",0
       db r_dx,f_16,"PSP seg",0
       db i_end
       db i_end
func27 db "random block read (FCB)",0
       db r_dsdx,f_fcb,"name",0
       db r_cx,f_16,"#records",0
       db i_end
       db r_ax,f_l8,"status",0
       db r_cx,f_16,"#read",0
       db i_end
func28 db "random block write (FCB)",0
       db r_dsdx,f_fcb,"name",0
       db r_cx,f_16,"#records",0
       db i_end
       db r_ax,f_l8,"status",0
       db r_cx,f_16,"#written",0
       db i_end
func29 db "parse filename to FCB",0
       db r_dssi,f_str0,"name",0
       db r_ax,f_l8,"options",0
       db i_end
       db r_ax,f_l8,"status",0
       db r_dsdx,f_fcb,"name",0
       db i_end
func2A db "get system date",0
       db i_end
       db r_ax,f_l8,"weekday",0
       db r_cx,f_16,"year",0
       db r_dx,f_h8,"month",0
       db r_dx,f_l8,"day",0
       db i_end
func2B db "set system date",0
       db r_cx,f_16,"year",0
       db r_dx,f_h8,"month",0
       db r_dx,f_l8,"day",0
       db i_end
       db r_ax,f_l8,"status",0
       db i_end
func2C db "get system time",0
       db i_end
       db r_cx,f_h8,"hour",0
       db r_cx,f_l8,"min",0
       db r_dx,f_h8,"sec",0
       db r_dx,f_l8,"1/100 sec",0
       db i_end
func2D db "set system time",0
       db r_cx,f_h8,"hour",0
       db r_cx,f_l8,"min",0
       db r_dx,f_h8,"sec",0
       db r_dx,f_l8,"1/100 sec",0
       db i_end
       db r_ax,f_l8,"status",0
       db i_end
func2E db "set verify flag",0
       db r_ax,f_l8,"verify",0
       db i_end
       db i_end
func2F db "get DTA address",0
       db i_end
       db r_esbx,f_32,"address",0
       db i_end
func30 db "get MS-DOS version number",0
       db i_end
       db r_ax,f_l8,"major",0
       db r_ax,f_h8,"minor",0
       db i_end
func31 db "terminate with return code and stay resident",0
       db r_ax,f_l8,"return code",0
       db r_dx,f_16,"#paragraphs",0
       db i_end
       db i_end
func32 db "get DPB pointer for a specified drive",0
       db r_dx,f_ch1,"drive",0
       db i_end
       db r_dsbx,f_strb,33,"DPB",0
       db i_end
func33 db "get/set <Ctrl><Break> flag",0
       db r_ax,f_l8,"get=0/set=1",0
       db i_end
       db r_dx,f_l8,"break",0
       db i_end
func34 db "get INDOS flag pointer",0
       db i_end
       db r_esbx,f_strb,1,"INDOS",0
       db i_end
func35 db "get interrupt vector",0
       db r_ax,f_l8,"int#",0
       db i_end
       db r_esbx,f_32,"address",0
       db i_end
func36 db "get free disk space",0
       db r_dx,f_ch1,"drive",0
       db i_end
       db r_ax,f_16,"sectors/cluster",0
       db r_cx,f_16,"bytes/sector",0
       db r_dx,f_16,"# of clusters",0
       db r_bx,f_16,"# of free clusters",0
       db i_end
func37 db "get/set code for command line switch",0
       db r_ax,f_l8,"get=0/set=1",0
       db i_end
       db r_dx,f_ch,"switch",0
       db r_cflag,f_bool,"status",0
       db i_end
func38 db "get/set country",0
       db i_end
       db r_cflag,f_bool,"status",0
       db i_end
func39 db "create subdirectory",0
       db r_dsdx,f_str0,"dir",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func3A db "remove subdirectory",0
       db r_dsdx,f_str0,"dir",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func3B db "set current directory",0
       db r_dsdx,f_str0,"dir",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func3C db "create file",0
       db r_dsdx,f_str0,"name",0
       db r_cx,f_16,"attrib",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func3D db "open file",0
       db r_dsdx,f_str0,"name",0
       db r_ax,f_l8,"mode",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"handle/code",0
       db i_end
func3E db "close file",0
       db r_bx,f_16,"handle",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func3F db "read from file",0
       db r_bx,f_16,"handle",0
       db r_cx,f_16,"#bytes",0
       db i_end
       db r_dsdx,f_stra,255,"buffer",0
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"bytes/code",0
       db i_end
func40 db "write to file",0
       db r_bx,f_16,"handle",0
       db r_cx,f_16,"#bytes",0
       db r_dsdx,f_stra,0,"buffer",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"bytes/code",0
       db i_end
func41 db "delete file",0
       db r_dsdx,f_str0,"name",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func42 db "move file pointer",0
       db r_bx,f_16,"handle",0
       db r_cxdx,f_32,"offset",0
       db r_ax,f_l8,"whence",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_dxax,f_32,"pos/code",0
       db i_end
func43 db "get/set file attributes",0
       db r_dsdx,f_str0,"name",0
       db r_ax,f_l8,"get=0/set=1",0
       db i_end
       db r_cx,f_16,"attrib",0
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func44 db "IOCTL function",0
       db r_ax,f_l8,"func#",0
       db r_bx,f_16,"handle/device/misc",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func45 db "duplicate handle",0
       db r_bx,f_16,"handle",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"new handle/code",0
       db i_end
func46 db "force duplicate of handle",0
       db r_bx,f_16,"handle",0
       db r_cx,f_16,"new handle",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func47 db "get current directory",0
       db r_dx,f_ch1,"drive",0
       db i_end
       db r_dssi,f_str0,"dir",0
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func48 db "allocate memory",0
       db r_bx,f_16,"#paragraphs",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"seg/code",0
       db r_bx,f_16,"#paragraphs free",0
       db i_end
func49 db "free memory",0
       db r_es,f_16,"seg",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func4A db "resize memory block",0
       db r_es,f_16,"seg",0
       db r_bx,f_16,"#paragraphs",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func4B db "execute program/overlay",0
       db r_ax,f_l8,"exec=0/overlay=3",0
       db r_dsdx,f_str0,"name",0
       db r_esbx,f_strb,20,"param block",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func4C db "terminate with return code",0
       db r_ax,f_l8,"return code",0
       db i_end
       db i_end
func4D db "get return code",0
       db i_end
       db r_ax,f_h8,"term. status",0
       db r_ax,f_l8,"return code",0
       db i_end
func4E db "search for first match",0
       db r_dsdx,f_str0,"mask",0
       db r_cx,f_16,"attrib",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func4F db "search for next match",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func50 db "set active PSP",0
       db r_bx,f_16,"PSP seg",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func51 db "get active PSP",0
       db i_end
       db r_bx,f_16,"PSP seg",0
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func52 db "get pointer to DIB",0
       db i_end
       db r_esbx,f_strb,71,"DIB",0
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func53 db "convert BPB to DPB",0
       db r_dssi,f_strb,36,"BPB",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func54 db "get verify flag",0
       db i_end
       db r_ax,f_l8,"verify",0
       db i_end
func55 db "create new PSP",0
       db r_cx,f_16,"old seg",0
       db r_dx,f_16,"new seg",0
       db i_end
       db r_cflag,f_bool,"status",0
       db i_end
func56 db "rename file",0
       db r_dsdx,f_str0,"old name",0
       db r_esdi,f_str0,"new name",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func57 db "get/set file date and time",0
       db r_ax,f_l8,"get=0/set=1",0
       db r_bx,f_16,"handle",0
       db i_end
       db r_cx,f_16,"packed time",0
       db r_dx,f_16,"packed date",0
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func58 db "get/set memory allocation strategy",0
       db r_ax,f_l8,"get=0/set=1",0
       db r_bx,f_16,"setting (for 1,3)",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"setting/code",0
       db i_end
func59 db "get extended error information",0
       db i_end
       db r_ax,f_16,"description",0
       db r_bx,f_h8,"cause",0
       db r_bx,f_l8,"action",0
       db r_cx,f_h8,"source",0
       db i_end
func5A db "create temporary file",0
       db r_dsdx,f_str0,"dir",0
       db r_cx,f_16,"attrib",0
       db i_end
       db r_dsdx,f_str0,"name",0
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func5B db "create new file",0
       db r_dsdx,f_str0,"name",0
       db r_cx,f_16,"attrib",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func5C db "(un)lock file record",0
       db r_ax,f_l8,"0=lock/1=unlock",0
       db r_bx,f_16,"handle",0
       db r_cxdx,f_32,"offset",0
       db r_sidi,f_32,"length",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func5D db "DOS internal/network call",0
       db i_end
       db i_end
func5E db "network function",0
       db i_end
       db i_end
func5F db "network redirection function",0
       db i_end
       db i_end
func60 db "canonicalize filename",0
       db r_dssi,f_str0,"old name",0
       db i_end
       db r_esdi,f_str0,"new name",0
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func61 db "reserved network function",0
       db i_end
       db i_end
func62 db "get PSP address",0
       db i_end
       db r_bx,f_16,"PSP seg",0
       db i_end
func63 db "DOS internal setting",0
       db i_end
       db i_end
func64 db "set device driver lookahead flag",0
       db i_end
       db i_end
func65 db "get extended country information",0
       db r_ax,f_l8,"func#",0
       db r_bx,f_16,"codepage",0
       db r_dx,f_16,"country ID",0
       db i_end
       db r_esdi,f_strb,41,"buffer",0
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func66 db "get/set code page",0
       db r_ax,f_l8,"1=get/2=set",0
       db i_end
       db r_bx,f_16,"active codepage",0
       db r_dx,f_16,"default codepage",0
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func67 db "set handle count",0
       db r_bx,f_16,"# of handles",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func68 db "commit file",0
       db r_bx,f_16,"handle",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func69 db "get/set disk serial number",0
       db i_end
       db i_end
func6A db "commit file",0
       db r_bx,f_16,"handle",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"code",0
       db i_end
func6B db "DOS internal function",0
       db i_end
       db i_end
func6C db "extended open",0
       db r_dssi,f_str0,"name",0
       db r_bx,f_16,"mode",0
       db r_cx,f_16,"attrib",0
       db r_dx,f_16,"reaction",0
       db i_end
       db r_cflag,f_bool,"status",0
       db r_ax,f_16,"handle/code",0
       db i_end
func6D db "find first ROM program",0
       db i_end
       db i_end
func6E db "find next ROM program",0
       db i_end
       db i_end
func6F db "DOS ROM function",0
       db i_end
       db i_end
func70 db "get/set internationalization info",0
       db i_end
       db i_end
func71 db "long filename function",0
       db i_end
       db i_end
func72 db "long filename findclose",0
       db i_end
       db i_end
func73 db "FAT32 function",0
       db i_end
       db i_end

@@stackstart:

program ends

end @@start
